/*
 * creates images from bullfrog's Map files
 * in BMP format
 *
 * Compiled normally, this file is a module without main().
 * Compiled with MAIN_XSIMMAP2 defined, it's a standalone program.
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "bulcommn.h"
#include "lbfileio.h"
#include "xsimmap2.h"


#ifdef MAIN_XSIMMAP2

int main (int argc, char **argv)
{
    printf("\nBullfrog Engine 2D Maps Extractor for Syndicate");
    printf("\n-------------------------------\n");
    if (argc<3)
    {
        printf("Not enought parameters.\n");
        printf("Usage:\n");
        printf("  %s <mapfile> <palfile>\n","xsimmap2");
        printf("Output: pic_map.bmp\n");
    	return 1;
    }

    printf("Loading map data...");
    MAPDATA *map=malloc(sizeof(MAPDATA));
    {
    int retcode;
    retcode=create_map_data(map, argv[1]);
    if (retcode!=0)
        switch (retcode)
        {
        case 1:
             printf("\nError: Cannot open map file: %s\n",argv[1]);
             return 1;
          case 3:
             printf("\nError: Map file is compressed with RNC, please decompress it\n");
             return 3;
        default:
             printf("\nError: Map data loading returned fail code %d\n",retcode);
             return 9;
        }
    }
    printf("Done.\n");

    printf("Loading palette ...");
    //Reading palette file
    char *palette=malloc(768);
    {
        char fname[255];
        sprintf (fname, "%s", argv[2]);
        int retcode=read_palette_rgb(palette,fname,256);
        if (retcode!=0)
          switch (retcode)
          {
          case 1:
             printf("\nError: Cannot open PAL file: %s\n",fname);
             return 11;
          case 2:
             printf("\nError: Cannot read %d colors from PAL file: %s\n",256,fname);
             return 12;
          case 3:
             printf("\nError: PAL file is compressed with RNC, please decompress it\n");
             return 13;
          default:
             printf("\nError: Loading palette file %s returned fail code %d\n",fname,retcode);
             return 19;
          }
    }
    printf("Done.\n");

    //Data loaded, now it's time for image generation
    printf("Generating map image...");
    int tile_size=4;

    //Full map size in pixels
    long full_map_w=map->max_x*tile_size;
    long full_map_h=map->max_y*tile_size;
    //if we don't need map endings, we can decrease size a little bit
    long mapwidth=full_map_w;//*4/5;
    long mapheight=full_map_h;//*7/8;
    char *image=malloc(mapwidth*mapheight);
    draw_simple_map_on_buffer_idx(image,map,mapwidth,mapheight,full_map_w/2,full_map_h/2,tile_size);
    printf("Done.\n");
    //Writting the bitmap
    char fname[20];
  	sprintf (fname, "pic_map.bmp");
    printf("Writing image to %s...",fname);
//   	write_bmp_24b (fname, mapwidth, mapheight,image);
    write_bmp_fn_idx(fname, mapwidth, mapheight, palette, image, 0, 1, 2, 4);
    //Finished
    free(image);
    free(palette);
    printf("Done.\n");
    free_map_data(map);
    free(map);
//    system("PAUSE");
    return 0;
}

#endif

void free_map_data(MAPDATA *map)
{
    if (map!=NULL)
        free(map->data);
}

int create_map_data(MAPDATA *map, char *map_fname)
// load the map file
{
    //Opening MAP file
    FILE *mapfp;
    mapfp = fopen (map_fname, "rb");
    if (!mapfp) return 1;
    //Checking for RNC compression
    if (rnc_compressed_file(mapfp)) return 3;
    map->max_x=read_int32_le_file(mapfp);
    map->max_y=read_int32_le_file(mapfp);
    map->max_z=read_int32_le_file(mapfp);
    map->data=malloc(map->max_x*map->max_y*map->max_z);
    long *look_up_table=malloc(sizeof(long)*map->max_x*map->max_y);
    // read in the header info
    int i;
    for( i=0; i < (map->max_x*map->max_y); i++ )
    {
        look_up_table[i]=read_int32_le_file(mapfp);
        // The firs 12 bytes seems not to count for offsets in file
        look_up_table[i] += 12;
    }
    // read the level data
    // populate the map
    int hidx;
    int widx;
    for( hidx=0; hidx < map->max_y; hidx++ )
        for( widx=0; widx < map->max_x; widx++ )
        {
            int tab_idx=(hidx*map->max_x) + widx;
            fseek(mapfp, look_up_table[tab_idx], SEEK_SET);
            fread (map->data + tab_idx*map->max_z, map->max_z, 1, mapfp);
        }
    fclose(mapfp);
    free(look_up_table);
    return 0;
}

int tile_is_crossed(unsigned char tilenum)
//Returns the shape that should be painted on
//a underlying tile. Also indicates if this tile
//should be skipped
{
    switch (tilenum)
    {
    case  34:return tcBottom;
    case 140:return tcRight;
    case 240:return tcTop;
    case 241:return tcLeft;
    case 242:return tcTop;
    case 243:return tcLeft;
    case 124:return tcSlash|tcBSlash;
    }
   return 0;
}

int tile_is_crossed_over(unsigned char tilenum)
//Returns the shape that should be painted on
//this tile. The tile is painted, whitch differs
//this function from tile_is_crossed(..)
{
    switch (tilenum)
    {
    case 100:return tcBottom|tcTop;
    case 101:return tcLeft|tcRight;
    case 102:return tcRight|tcTop;
    case 103:return tcRight|tcBottom;
    case 104:return tcLeft|tcBottom;
    case 105:return tcLeft|tcTop;
    case 114:return tcBottom;
    case 115:return tcRight;
    case 116:return tcTop;
    case 117:return tcLeft;
    case 118:return tfBottom|tfRight|tfLeft|tfBottom;
    }
   return 0;
}

int tile_is_wire_fence(unsigned char tilenum)
{
    switch (tilenum)
    {
    case 130:return tfLeft;
    case 131:return tfTop;
    case 132:return tfBottom;
    case 133:return tfRight;
    case 134:return tfLeft|tfTop;
    case 135:return tfTop|tfRight;
    case 136:return tfLeft|tfBottom;
    case 137:return tfBottom|tfRight;
    case 138:return tfLeft|tfRight;
    case 139:return tfTop|tfBottom;
    case 141:return tfRight;
    case 142:return tfLeft;
    case 143:return tfLeft;
    case 144:return tfBottom;
    case 146:return tfTop;
    case 147:return tfTop;
    case 148:return tfRight;
    case 150:return tfLeft;
    case 151:return tfLeft;
    case 152:return tfBottom;
    case 154:return tfTop;
    case 155:return tfTop;
    }
   return 0;
}

unsigned char tile_fence_color_idx(unsigned char tilenum)
{
    switch (tilenum)
    {
    case 100:
    case 101:
    case 102:
    case 103:
    case 104:
    case 105:
    case 114:
    case 115:
    case 116:
    case 117:return 225;
    case  34:
    case 140:
    case 240:
    case 241:
    case 242:
    case 243:return 7;
    case 118:return 107;
    }
   return 16;
}

int draw_simpletile_on_buffer_idx(unsigned char *buffer,int width,int height,int wpos,int hpos,int tile_idx,int tile_fence,unsigned char fence_color,int tile_size)
//Draws a simple representation of tile number tile_idx to buffer (whitch has width*height bytes),
// starting (top left corner) at (wpos,hpos).
{
    int widx;
    int hidx;
    for (widx=0;widx<tile_size;widx++)
      for (hidx=0;hidx<tile_size;hidx++)
      {
          int buf_h=hpos+hidx;
          int buf_w=wpos+widx;
          if ((buf_h>=0)&&(buf_h<height)&&(buf_w>=0)&&(buf_w<width))
          {
              int buf_idx=buf_h*width + buf_w;
              int til_idx=hidx*tile_size + widx;
              buffer[buf_idx]=(unsigned char)(tile_idx%256);
          }
      }
    for (widx=0;widx<tile_size;widx++)
    {
        int buf_w=wpos+widx;
        int buf_h=hpos;
        int buf_idx=buf_h*width + buf_w;
        if (tile_fence&tfTop)
            buffer[buf_idx]=fence_color;
        buf_h=hpos+tile_size-1;
        buf_idx=buf_h*width + buf_w;
        if (tile_fence&tfBottom)
            buffer[buf_idx]=fence_color;
    }
    for (hidx=0;hidx<tile_size;hidx++)
    {
        int buf_h=hpos+hidx;
        int buf_w=wpos;
        int buf_idx=buf_h*width + buf_w;
        if (tile_fence&tfLeft)
            buffer[buf_idx]=fence_color;
        buf_w=wpos+tile_size-1;
        buf_idx=buf_h*width + buf_w;
        if (tile_fence&tfRight)
            buffer[buf_idx]=fence_color;
    }
    if (tile_fence&tcLeft)
    {
        int buf_h=hpos+tile_size/2;
        if (!(tile_size%2)) buf_h--;
        for (widx=0;widx<tile_size/2;widx++)
        {
            int buf_w=wpos+widx;
            int buf_idx=buf_h*width + buf_w;
            buffer[buf_idx]=fence_color;
        }
    }
    if (tile_fence&tcRight)
    {
        int wid_start=tile_size/2;
        int buf_h=hpos+tile_size/2;
        if (tile_size%2) wid_start++;
        for (widx=wid_start;widx<tile_size;widx++)
        {
            int buf_w=wpos+widx;
            int buf_idx=buf_h*width + buf_w;
            buffer[buf_idx]=fence_color;
        }
    }
    if (tile_fence&tcBottom)
    {
        int heig_start=tile_size/2;
        int buf_w=wpos+tile_size/2;
        if (tile_size%2) heig_start++;
        else buf_w--;
        for (hidx=heig_start;hidx<tile_size;hidx++)
        {
            int buf_h=hpos+hidx;
            int buf_idx=buf_h*width + buf_w;
            buffer[buf_idx]=fence_color;
        }
    }
    if (tile_fence&tcTop)
    {
        int buf_w=wpos+tile_size/2;
        for (hidx=0;hidx<tile_size/2;hidx++)
        {
            int buf_h=hpos+hidx;
            int buf_idx=buf_h*width + buf_w;
            buffer[buf_idx]=fence_color;
        }
    }
    if (tile_fence&tcSlash)
    {
        for (hidx=0;hidx<tile_size;hidx++)
        {
            int buf_w=wpos+hidx;
            int buf_h=hpos+hidx;
            int buf_idx=buf_h*width + buf_w;
            buffer[buf_idx]=fence_color;
        }
    }
    if (tile_fence&tcBSlash)
    {
      if (tile_size>2)
        for (hidx=0;hidx<tile_size;hidx++)
        {
          float distnce=(float)(hidx)+0.5f-((float)tile_size)/2;
        if ((distnce>0.6f)||(distnce<-0.6f))
          {
            int buf_w=wpos+tile_size-hidx-1;
            int buf_h=hpos+hidx;
            int buf_idx=buf_h*width + buf_w;
            buffer[buf_idx]=fence_color;
          }
        }
    }

    return 0;
}


void draw_simple_map_on_buffer_idx(unsigned char *buffer,MAPDATA *map,int width,int height,int wpos,int hpos,int tile_size)
// writes the map to a memory buffer in indexed form
//wpos and hpos are coordinates of the centre of the displayed map
{
    //Processed map coordinate range
    //TODO: FIX VALUES for partial loading!!!
    int map_start_y=0;
    int map_end_y=map->max_y;
    int map_start_x=0;
    int map_end_x=map->max_x;
    long full_map_w=map->max_x*tile_size;
    long full_map_h=map->max_y*tile_size;
    //Now going through tiles and creating map
    int map_x;
    for (map_x=map_start_x;map_x<map_end_x;map_x++)
    {
        int map_y;
        for (map_y=map_start_y;map_y<map_end_y;map_y++)
        {
            int screen_w = wpos -  width/2 -  (full_map_w-width)/2 + map_x*tile_size;
            int screen_h = hpos - height/2 - (full_map_h-height)/2 + map_y*tile_size;
            int map_idx=((map_y*map->max_x) + map_x)*map->max_z;
            int map_z=map->max_z;
            int tile_idx;
            int tile_fence=0;
            unsigned char fence_color=0;
            int tile_fence_add=0;
            //Getting the topmost tile
            do
            {
                map_z--;
                tile_idx=map->data[map_idx + map_z];
                tile_fence_add=tile_is_wire_fence(tile_idx);
                tile_fence_add|=tile_is_crossed(tile_idx);
                if (tile_fence_add)
                {
                    tile_fence|=tile_fence_add;
                    fence_color=tile_fence_color_idx(tile_idx);
                }
            }
            while ((map_z>0)&&((tile_idx<5)||(tile_idx==124)||(tile_fence_add)));
            tile_fence_add=tile_is_crossed_over(tile_idx);
            if (tile_fence_add)
            {
                tile_fence|=tile_fence_add;
                fence_color=tile_fence_color_idx(tile_idx);
            }
            //Now drawing tile on simple map
            draw_simpletile_on_buffer_idx(buffer,width,height,screen_w,screen_h,tile_idx,tile_fence,fence_color,tile_size);
        }
    }
}

void draw_simple_map_on_buffer_rgb(unsigned char *buffer,MAPDATA *map,int width,int height,int wpos,int hpos,int scale)
//Writes the map to a memory buffer in red-green-blue form,
// wpos and hpos are coordinates of the centre of the displayed map.
{
}

