 /*
 * converts picture data from bullfrog's Buildings Sprites files
 * to a BMP images
 *
 * Compiled normally, this file is a module without main().
 * Compiled with MAIN_XSPRBLK defined, it's a standalone program.
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "bulcommn.h"
#include "lbfileio.h"
#include "xsprblk.h"

#ifdef MAIN_XSPRBLK

int main (int argc, char **argv)
{
    printf("\nBullfrog Engine 2D Buildings Sprites Extractor for Map graphic");
    printf("\n(probably works only with Syndicate HBLK01.DAT)");
    printf("\n-------------------------------\n");
    if (argc<3)
    {
        printf("Not enought parameters.\n");
        printf("Usage:\n");
        printf("  %s <datafile> <palfile>\n","xsprblk");
        printf("Output: pic????.bmp\n");
    	return 1;
    }
    SPRITETILE stile;
    //Opening and reading DAT file
    {
        FILE *datfp;
        char fname[255];
        sprintf (fname, "%s", argv[1]);
        datfp = fopen (fname, "rb");
        if (!datfp)
        {
            printf("Error: Can't open DAT file %s.\n", fname);
        	return 1;
        }
        stile.datasize=file_length_opened(datfp);
        stile.data=malloc(stile.datasize+4);
        fseek(datfp, 0, SEEK_SET);
        fread (stile.data, stile.datasize, 1, datfp);
        fclose(datfp);
    }
    printf("Loading palette ...");
    //Reading palette file
    char *palette=malloc(256*3);
    {
        char fname[255];
        sprintf (fname, "%s", argv[2]);
        int retcode=read_palette_rgb(palette,fname,256);
        if (retcode!=0)
          switch (retcode)
          {
          case 1:
             printf("\nError: Cannot open PAL file: %s\n",fname);
             return 11;
          case 2:
             printf("\nError: Cannot read %d colors from PAL file: %s\n",256,fname);
             return 12;
          case 3:
             printf("\nError: PAL file is compressed with RNC, please decompress it\n");
             return 13;
          default:
             printf("\nError: Loading palette file %s returned fail code %d\n",fname,retcode);
             return 19;
          }
    }
    printf("Done.\n");

    //Filling structure with information about Tile file content
    stile.subwidth   = 32;
    stile.subheight  = 16;
    stile.subcount_w = 2;
    stile.subcount_h = 3;
    stile.num_tiles  = 256;
    long num_subtiles=stile.num_tiles*stile.subcount_w*stile.subcount_h;
    int tile_size=stile.subwidth*stile.subcount_w*stile.subheight*stile.subcount_h;
    //checking if the Tile file size is right
    //It is hard to say how large the file should be, because every tile
    //is made of subtiles, and each subtile may be used in many tiles.
    unsigned long proper_stile_size=num_subtiles*4 + num_subtiles*
          (stile.subwidth*stile.subheight * 5/8);
    if (stile.datasize<(proper_stile_size/3))
    {
        printf("Warning: DAT file is surprisingly small!\n");
    }
    //Checking for RNC compression
    if (rnc_compressed_buf(stile.data))
    {
        printf("Warning: DAT file seems to be RNC-compressed!\n");
        printf("         Decompress it first with DERNC!\n");
    }
    stile.look_up_table=malloc(sizeof(long)*num_subtiles);
    // Reading Tile file header info
    // (an array of the offsets of all the subtiles)
    {
        int i;
        for( i=0; i < num_subtiles; i++ )
            stile.look_up_table[i]=read_int32_le_buf(stile.data+4*i);
    }

    int processed=4*num_subtiles;
    //Interpreting DAT file - Looping through tiles
    unsigned char *tile_pix = malloc(tile_size);
    unsigned char *tile_alpha = malloc(tile_size);
    int picnum;
    for (picnum=0; picnum<stile.num_tiles; picnum++)
    {
    	printf("\rExtracting: tile number %*d", 4, picnum);
        processed+=get_sprite_tile(tile_pix, tile_alpha, picnum, stile);
        char fname[20];
    	sprintf(fname, "pic%0*d.bmp", 4, picnum);
       	write_bmp_fn_idx(fname, stile.subwidth*stile.subcount_w, stile.subheight*stile.subcount_h,
            palette, tile_pix, 0, 1, 2, 4);
    }
    free(tile_pix);
    free(tile_alpha);
    //Summary informations
    printf("\nExtraction complete.\n");
    printf("Processed %ld of %ld bytes of DAT file.\n",(long)processed,(long)stile.datasize);
    int unused=stile.datasize-processed;
    if (unused>=0)
        printf("Bytes skipped: %d\n",unused);
      else
        printf("Bytes overlapping: %d\n",-unused);
    free(stile.look_up_table);
    free(stile.data);
    return 0;
}

#endif

void free_map_tiles_rawlist(MAPTILELIST *list)
{
    int picnum;
    for (picnum=0; picnum < list->num_tiles; picnum++)
    {
    free(list->tiles[picnum].data);
    free(list->tiles[picnum].alpha);
    }
    list->num_tiles=0;
    free(list->palette);
//    free(list->tiles);
    list->tiles=NULL;
}

/**
 * Reads all the titles to MAPTILELIST structure from given files.
 * @return On success, gives 0.
 */
int create_map_tiles_rawlist(MAPTILELIST *list,const char *sprite_file_name,const char *pal_file_name)
{
    SPRITETILE stile;
    //printf("    Opening and reading DAT file\n");
    {
        FILE *datfp;
        datfp = fopen (sprite_file_name, "rb");
        if (!datfp)
        {
        	return SPRBLK_CANNOT_OPEN_DAT;
        }
        stile.datasize = file_length_opened(datfp);
        stile.data = malloc(stile.datasize+4);
        if (stile.data == NULL)
        {
            fclose(datfp);
            return SPRBLK_MALLOC_ERROR;
        }
        fseek(datfp, 0, SEEK_SET);
        fread (stile.data, stile.datasize, 1, datfp);
        fclose(datfp);
    }
    //printf("    Reading palette file\n");
    {
        list->palette=malloc(256*3);
        int retcode=read_palette_rgb(list->palette,pal_file_name,256);
        if (retcode!=0)
          switch (retcode)
          {
          case 1:
             return SPRBLK_CANNOT_OPEN_PAL;
          default:
             return SPRBLK_CONTENT_BAD_PAL;
          }
    }
    //printf("    Filling internal structure with information about Tile file content\n");
    stile.subwidth   = 32;
    stile.subheight  = 16;
    stile.subcount_w = 2;
    stile.subcount_h = 3;
    stile.num_tiles  = 256;
    long num_subtiles=stile.num_tiles*stile.subcount_w*stile.subcount_h;
    int tile_size=stile.subwidth*stile.subcount_w*stile.subheight*stile.subcount_h;
    //checking if the Tile file size is right
    //It is hard to say how large the file should be, because every tile
    //is made of subtiles, and each subtile may be used in many tiles.
    unsigned long proper_stile_size=num_subtiles*4 + num_subtiles*
          (stile.subwidth*stile.subheight * 5/8);
    if (stile.datasize<(proper_stile_size/3))
             return SPRBLK_CONTENT_BAD_DAT;
    stile.look_up_table=malloc(sizeof(long)*num_subtiles);
    //printf("    Reading Tile file header info\n");
    // (an array of the offsets of all the subtiles)
    {
        int i;
        for( i=0; i < num_subtiles; i++ )
            stile.look_up_table[i]=read_int32_le_buf(stile.data+4*i);
    }
    int processed=4*num_subtiles;
    //printf("    Setting output structure informations\n");
    list->num_tiles = stile.num_tiles;
    list->tiles = malloc(list->num_tiles*sizeof(MAPTILE));
    if (list->tiles == NULL)
        return SPRBLK_MALLOC_ERROR;
    //Interpreting DAT file - Looping through tiles
    int picnum;
    for (picnum=0; picnum<list->num_tiles; picnum++)
    {
        //printf("    Interpreting DAT tile %d\n",picnum);
        unsigned char *tile_pix = malloc(tile_size);
        unsigned char *tile_alpha = malloc(tile_size);
        if ( (tile_pix == NULL) || (tile_alpha == NULL) )
        {
            free(tile_pix); free(tile_alpha); free(list->tiles);
            list->tiles = NULL;
            free(stile.look_up_table);
            free(stile.data);
            return SPRBLK_MALLOC_ERROR;
        }
        list->tiles[picnum].width = stile.subwidth*stile.subcount_w;
        list->tiles[picnum].height = stile.subheight*stile.subcount_h;
        list->tiles[picnum].alpha = tile_alpha;
        list->tiles[picnum].data = tile_pix;
        processed += get_sprite_tile(tile_pix, tile_alpha, picnum, stile);
    }
    //Freeing temporary structures
    free(stile.look_up_table);
    free(stile.data);
    return 0;
}

/** Extracts the whole tile from building sprite file loaded into
 * SPRITETILE structure. Fills pix_map and alpha_map tables.
 * Requires tile number (usually 0..255). Tile must consist of 6 subtiles.
 */
int get_sprite_tile(unsigned char *pix_map,unsigned char *alpha_map, int num, SPRITETILE src)
{
    //Number of source buffer bytes processed
    int processed=0;
    //Subtiles count
    long num_subtiles=src.num_tiles*src.subcount_w*src.subcount_h;
    //Title consists of 6 subtitles indexed 0..5.
    unsigned char *subtile_pix = malloc(src.subwidth * src.subheight);
    unsigned char *subtile_alpha = malloc(src.subwidth * src.subheight);
    if ( (subtile_pix == NULL) || (subtile_alpha == NULL) )
    {
        free(subtile_pix);
        free(subtile_alpha);
        return processed;
    }
    //First three are left half of image, next three are right half
    //So they come as:
    //    |-----------------|
    //    |   0    |   3    |
    //    |-----------------|
    //    |   1    |   4    |
    //    |-----------------|
    //    |   2    |   5    |
    //    |-----------------|
    // Now going through subtitles
    int subnum;
    for (subnum=0;subnum<src.subcount_w*src.subcount_h;subnum++)
    {
        int offset=src.look_up_table[(num*src.subcount_w*src.subcount_h)+subnum];
        if (offset >= 4*num_subtiles)
            processed+=get_sprite_subtile(subtile_pix,subtile_alpha,
                                          &src.data[offset],src.datasize-offset,
                                          src.subwidth,src.subheight);
         else
            fill_sprite_subtile(subtile_pix,subtile_alpha,src.subwidth,src.subheight,0);
        int h,w;
        int sub_w0=src.subwidth *(subnum/src.subcount_h);//0 for subnum<3, 1 for subnum<6
        int sub_h0=src.subheight*(subnum%src.subcount_h);//usually 0, 16 or 32
        for( h=0; h < src.subheight; h++ )
        for( w=0; w < src.subwidth; w++ )
        {
            int map_idx=w+sub_w0 + ((h+sub_h0)*src.subwidth*src.subcount_w);
            pix_map  [map_idx] = subtile_pix  [w + (h*src.subwidth)];
            alpha_map[map_idx] = subtile_alpha[w + (h*src.subwidth)];
        }
    }
    free(subtile_pix);
    free(subtile_alpha);
    return processed;
}

void fill_sprite_subtile(unsigned char *pix_map,unsigned char *alpha_map, int width, int height, int hstart)
{
    int widx;
    int hidx;
    for( hidx=hstart; hidx < height; hidx++ )
    {
        int hpos=hidx*width;
        for( widx=0; widx < width; widx++ )
            {
                pix_map[hpos+widx] = 255;
                alpha_map[hpos+widx]=255;
            }
    }
}

/** Give it the buffer with subtile from the hblk file, and it will
 * fill the arrays with the image data. Returns number of source buffer bytes processed.
 */
int get_sprite_subtile(unsigned char *pix_map,unsigned char *alpha_map, unsigned char *src_buf, int src_bufsize, int width, int height)
{
    int processed=0;
    // Determining the chunk size (usually 20)
    int chunk_size=(width/8)*5;
    if ((width%8)!=0) chunk_size+=5;
    // for every horizontal strip of pict...
    int hidx;
    for( hidx=0; hidx < height; hidx++ )
        {
            //If the buffer is truncated, fill the rest of output
            if (src_bufsize < processed+chunk_size)
            {
                fill_sprite_subtile(pix_map,alpha_map,width,height,hidx);
                return processed;
            }
            int hpos=hidx*width;
            processed+=chunk_size;
            unsigned char *dat_chunk=src_buf+chunk_size*hidx;
            int widx;
            // On error, do nothing
            if ( (pix_map == NULL) || (alpha_map == NULL) || (src_buf == NULL) )
                continue;
            for( widx=0; widx < width; widx++ )
            {
                // check for alpha and fill the buffer
                if( nth_bit_fourbytes( dat_chunk, 31-widx ) )
                {
                    alpha_map[hpos+widx]=255;
                    pix_map[hpos+widx]=255;
                }
                 else
                {
                    alpha_map[hpos+widx]=0;
                    pix_map[hpos+widx] = ( nth_bit( dat_chunk[4  + (widx/8)], 7 - (widx%8) ) * 1 )
                                   + ( nth_bit( dat_chunk[8  + (widx/8)], 7 - (widx%8) ) * 2 )
                                   + ( nth_bit( dat_chunk[12 + (widx/8)], 7 - (widx%8) ) * 4 )
                                   + ( nth_bit( dat_chunk[16 + (widx/8)], 7 - (widx%8) ) * 8 );
                }
            }
        }


    return processed;
}
